@@ -26,7 +26,7 @@ Fetch options:
     -c, --new-column <name>    Put the fetched values in a new column instead.
     --jql <selector>           Apply jql selector to API returned JSON value.
     --rate-limit <qps>         Rate Limit in Queries Per Second. [default: 5]
-    --http-header <k:v>        Pass custom header(s) to the server.
+    --http-header <key:value>  Pass custom header(s) to the server.
     --store-error              On error, store error code/message instead of blank value.
     --cookies                  Allow cookies.
 
@@ -152,9 +152,7 @@ pub fn run(argv: &[&str]) -> CliResult<()> {
 
     #[allow(unused_assignments)]
     let mut record = csv::ByteRecord::new();
-    for row in rdr.byte_records() {
-        record = row?;
-
+    while rdr.read_byte_record(&mut record)? {
         if !args.flag_quiet {
             progress.inc(1);
         }
