@@ -6,10 +6,12 @@ use anyhow::{anyhow, Result};
 use csv::ByteRecord;
 use indicatif::{ProgressBar, ProgressDrawTarget};
 use jsonschema::{output::BasicOutput, JSONSchema};
+#[allow(unused_imports)]
 use log::{debug, info};
+use rayon::prelude::*;
 use serde::Deserialize;
 use serde_json::{value::Number, Map, Value};
-use std::{env, fs::File, io::BufReader, io::BufWriter, io::Read, io::Write, ops::Add};
+use std::{env, fs::File, io::BufReader, io::BufWriter, io::Read, io::Write, str};
 
 macro_rules! fail {
     ($mesg:expr) => {
@@ -17,8 +19,11 @@ macro_rules! fail {
     };
 }
 
+// number of CSV rows to process in a batch
+const BATCH_SIZE: usize = 16000;
+
 static USAGE: &str = "
-Validate CSV data with JSON Schema, and put invalid records into separate file.
+Validate CSV data with JSON Schema, and put invalid records into a separate file.
 
 Example output files from `mydata.csv`. If piped from stdin, then filename is `stdin.csv`.
 
@@ -26,9 +31,9 @@ Example output files from `mydata.csv`. If piped from stdin, then filename is `s
 * mydata.csv.invalid
 * mydata.csv.validation-errors.jsonl
 
-JSON Schema can be a local file or a URL. 
+JSON Schema can be a local file or a URL.
 
-When run without JSON Schema, only a simple CSV check (RFC 4180) is performed, with the caveat that 
+When run without JSON Schema, only a simple CSV check (RFC 4180) is performed, with the caveat that
  on non-Windows machines, each record is delimited by a LF (\\n) instead of CRLF (\\r\\n).
 
 
@@ -52,7 +57,7 @@ Common options:
     -q, --quiet                Don't show progress bars.
 ";
 
-#[derive(Deserialize, Debug)]
+#[derive(Deserialize)]
 struct Args {
     flag_fail_fast: bool,
     flag_valid: Option<String>,
@@ -67,47 +72,45 @@ struct Args {
 pub fn run(argv: &[&str]) -> CliResult<()> {
     let args: Args = util::get_args(USAGE, argv)?;
 
-    // dbg!(&args);
-
-    let rconfig = Config::new(&args.arg_input)
+    let mut rconfig = Config::new(&args.arg_input)
         .delimiter(args.flag_delimiter)
         .no_headers(args.flag_no_headers);
 
     let mut rdr = rconfig.reader()?;
 
+    // if no json schema supplied, only let csv reader validate csv file
+    if args.arg_json_schema.is_none() {
+        // just read csv file and let csv reader report problems
+        let mut record = csv::ByteRecord::new();
+        while rdr.read_byte_record(&mut record)? {
+            // this loop is for csv::reader to do basic csv validation on read
+        }
+
+        let msg = "Can't validate without schema, but csv looks good.".to_string();
+        info!("{msg}");
+        println!("{msg}");
+
+        return Ok(());
+    }
+
     let headers = rdr.byte_headers()?.clone();
 
-    let input_path: &str = &args.arg_input.unwrap_or_else(|| "stdin.csv".to_string());
-
-    let valid_suffix: &str = &args.flag_valid.unwrap_or_else(|| "valid".to_string());
-    let mut valid_wtr = Config::new(&Some(input_path.to_owned() + "." + valid_suffix)).writer()?;
-
-    let invalid_suffix: &str = &args.flag_invalid.unwrap_or_else(|| "invalid".to_string());
-    let mut invalid_wtr =
-        Config::new(&Some(input_path.to_owned() + "." + invalid_suffix)).writer()?;
-
-    let wtr_capacitys = env::var("QSV_WTR_BUFFER_CAPACITY")
-        .unwrap_or_else(|_| DEFAULT_WTR_BUFFER_CAPACITY.to_string());
-    let wtr_buffer: usize = wtr_capacitys.parse().unwrap_or(DEFAULT_WTR_BUFFER_CAPACITY);
-
-    let mut error_report_file = BufWriter::with_capacity(
-        wtr_buffer,
-        File::create(input_path.to_owned() + ".validation-errors.jsonl")
-            .expect("unable to create error report file"),
-    );
-
     // prep progress bar
     let progress = ProgressBar::new(0);
+    // for purpose of full file row count, prevent CSV reader to abort on incosistent column count
+    rconfig = rconfig.flexible(true);
+    let record_count = util::count_rows(&rconfig);
+    rconfig = rconfig.flexible(false);
+
     if !args.flag_quiet {
-        let record_count = util::count_rows(&rconfig.flexible(true));
         util::prep_progress(&progress, record_count);
     } else {
         progress.set_draw_target(ProgressDrawTarget::hidden());
     }
 
-    // check if need to validate via json schema, or just let csv reader validate csv file
-    if let Some(json_schema_uri) = &args.arg_json_schema {
-        let (schema_json, schema_compiled): (Value, JSONSchema) = match load_json(json_schema_uri) {
+    // parse and compile supplied JSON Schema
+    let (schema_json, schema_compiled): (Value, JSONSchema) =
+        match load_json(&args.arg_json_schema.unwrap()) {
             Ok(s) => {
                 // parse JSON string
                 match serde_json::from_str(&s) {
@@ -121,7 +124,6 @@ pub fn run(argv: &[&str]) -> CliResult<()> {
                         }
                     }
                     Err(e) => {
-                        //error!("Unable to parse schema json. error: {}", e);
                         return fail!(format!("Unable to parse schema json. error: {e}"));
                     }
                 }
@@ -130,143 +132,143 @@ pub fn run(argv: &[&str]) -> CliResult<()> {
                 return fail!(format!("Unable to retrieve json. error: {e}"));
             }
         };
-        debug!("compiled schema: {:?}", &schema_compiled);
 
-        let mut valid_file_empty: bool = true;
-        let mut invalid_file_empty: bool = true;
+    // debug!("compiled schema: {:?}", &schema_compiled);
 
-        // amortize memory allocation by reusing record
-        #[allow(unused_assignments)]
-        let mut record = csv::ByteRecord::new();
+    // how many rows read and processed as batches
+    let mut row_number: usize = 0;
+    // how many invalid rows found
+    let mut invalid_count: usize = 0;
 
-        let mut row_index: u32 = 0;
-        let mut invalid_count: u32 = 0;
+    // amortize memory allocation by reusing record
+    #[allow(unused_assignments)]
+    let mut record = csv::ByteRecord::new();
+    // reuse batch buffer
+    let mut batch = Vec::with_capacity(BATCH_SIZE);
+    let mut valid_flags: Vec<bool> = Vec::with_capacity(record_count as usize);
+    let mut validation_error_messages: Vec<String> = Vec::new();
 
-        while rdr.read_byte_record(&mut record)? {
-            row_index = row_index.add(1);
-
-            let instance: Value = match to_json_instance(&headers, &record, &schema_json) {
-                Ok(obj) => obj,
-                Err(e) => {
-                    return fail!(format!(
-                        "Unable to convert CSV to json. row: {row_index}, error: {e}"
-                    ));
-                }
-            };
-
-            debug!("instance[{}]: {:?}", &row_index, &instance);
-
-            match validate_json_instance(&instance, &schema_compiled) {
-                Ok(validation_result) => {
-                    let results = &validation_result["valid"];
-
-                    debug!("validation[{row_index}]: {results:?}");
-
-                    let valid_flag = match results.as_bool().to_owned() {
-                        Some(b) => b,
-                        None => {
-                            return fail!(format!(
-                                "Unexpected validation result. row: {row_index}, result: {results:?}"
-                            ));
-                        }
-                    };
-
-                    match valid_flag {
-                        true => {
-                            if valid_file_empty {
-                                valid_wtr.write_byte_record(&headers)?;
-                                valid_file_empty = false;
-                            }
-
-                            valid_wtr.write_byte_record(&record)?;
-                        }
-                        false => {
-                            invalid_count = invalid_count.add(1);
-
-                            debug!(
-                                "schema violation. row: {row_index}, violation: {validation_result:?}"
-                            );
-                            // dbg!(&validation_result, &record);
-
-                            // write to invalid file
-
-                            if invalid_file_empty {
-                                invalid_wtr.write_byte_record(&headers)?;
-                                invalid_file_empty = false;
-                            }
-
-                            invalid_wtr.write_byte_record(&record)?;
-
-                            // write to error report
-                            let mut enriched_results_map = validation_result
-                                .as_object()
-                                .expect("get validation results as map")
-                                .clone();
-                            let _ = enriched_results_map.insert(
-                                "row_index".to_string(),
-                                Value::Number(Number::from(row_index)),
-                            );
-                            let enriched_results: Value = Value::Object(enriched_results_map);
-
-                            error_report_file
-                                .write_all(format!("{enriched_results}\n").as_bytes())
-                                .expect("unable to write to validation error report");
-
-                            // for fail-fast, just break out of loop
-                            if args.flag_fail_fast {
-                                break;
-                            }
-                        }
+    // main loop to read CSV and construct batches for parallel processing.
+    // each batch is processed via Rayon parallel iterator.
+    // loop exits when batch is empty.
+    loop {
+        for _ in 0..BATCH_SIZE {
+            match rdr.read_byte_record(&mut record) {
+                Ok(has_data) => {
+                    if has_data {
+                        row_number += 1;
+                        record.push_field(row_number.to_string().as_bytes());
+                        batch.push(record.to_owned());
+                    } else {
+                        // nothing else to add to batch
+                        break;
                     }
                 }
                 Err(e) => {
-                    return fail!(format!("Unable to validate. row: {row_index}, error: {e}"));
+                    return Err(CliError::Other(format!(
+                        "Error reading row: {row_number}: {e}"
+                    )));
                 }
             }
+        }
 
-            if !args.flag_quiet {
-                progress.inc(1);
+        if batch.is_empty() {
+            // break out of infinite loop when at EOF
+            break;
+        }
+
+        let batch_size = batch.len();
+
+        // do actual validation via Rayon parallel iterator
+        let validation_results: Vec<Option<String>> = batch
+            .par_iter()
+            .map(|record| {
+                match do_json_validation(&headers, record, &schema_json, &schema_compiled) {
+                    Ok(o) => o,
+                    Err(e) => {
+                        panic!("Unrecoverable error: {e:?}");
+                    }
+                }
+            })
+            .collect();
+
+        batch.clear();
+
+        // write to validation error report, but keep Vec<bool> to gen valid/invalid files later
+        // because Rayon collect() guaranteeds original order, can sequentially append results to vector with each batch
+        for result in validation_results {
+            match result {
+                Some(validation_error_msg) => {
+                    invalid_count += 1;
+                    valid_flags.push(false);
+
+                    validation_error_messages.push(validation_error_msg);
+                }
+                None => {
+                    valid_flags.push(true);
+                }
             }
-        } // end main while loop over csv records
-
-        // flush error report; file gets closed automagically when out-of-scope
-        error_report_file.flush().unwrap();
-
-        use thousands::Separable;
+        }
 
         if !args.flag_quiet {
-            progress.set_message(format!(
-                " validated {} records.",
-                progress.length().separate_with_commas()
-            ));
-            util::finish_progress(&progress);
+            progress.inc(batch_size as u64);
         }
 
-        // done with validation; print output
-        if args.flag_fail_fast {
-            let msg = format!(
-                "fail-fast enabled. stopping after first invalid record at row {}",
-                row_index.separate_with_commas()
-            );
-            info!("{msg}");
-            println!("{msg}");
-        } else {
-            let msg = format!(
-                "{} out of {} records invalid.",
-                invalid_count.separate_with_commas(),
-                row_index.separate_with_commas()
-            );
-            info!("{msg}");
-            println!("{msg}");
+        // for fail-fast, exit loop if batch has any error
+        if args.flag_fail_fast && invalid_count > 0 {
+            break;
         }
+    } // end infinite loop
+
+    // only write out invalid/valid/errors output files if there are actually invalid records.
+    // if 100% invalid, then valid file is not needed. but this is rare so live with creating empty file.
+    if invalid_count > 0 {
+        let input_path = args
+            .arg_input
+            .clone()
+            .unwrap_or_else(|| "stdin.csv".to_string());
+
+        write_error_report(&input_path, validation_error_messages)?;
+
+        let valid_suffix = args.flag_valid.unwrap_or_else(|| "valid".to_string());
+        let invalid_suffix = args.flag_invalid.unwrap_or_else(|| "invalid".to_string());
+
+        split_invalid_records(
+            &rconfig,
+            &valid_flags[..],
+            &headers,
+            &input_path,
+            &valid_suffix,
+            &invalid_suffix,
+        )?;
+    }
+
+    use thousands::Separable;
+
+    if !args.flag_quiet {
+        progress.set_message(format!(
+            " validated {} records.",
+            progress.length().separate_with_commas()
+        ));
+        util::finish_progress(&progress);
+    }
+
+    // done with validation; print output
+    if args.flag_fail_fast {
+        let msg = format!(
+            "fail-fast enabled. stopped after row {}.\n{} out of {} records invalid.",
+            row_number.separate_with_commas(),
+            invalid_count.separate_with_commas(),
+            row_number.separate_with_commas()
+        );
+        info!("{msg}");
+        println!("{msg}");
     } else {
-        // just read csv file and let csv reader report problems
-        let mut record = csv::ByteRecord::new();
-        while rdr.read_byte_record(&mut record)? {
-            // this loop is for csv::reader to do basic csv validation on read
-        }
-
-        let msg = "Can't validate without schema, but csv looks good.".to_string();
+        let msg = format!(
+            "{} out of {} records invalid.",
+            invalid_count.separate_with_commas(),
+            row_number.separate_with_commas()
+        );
         info!("{msg}");
         println!("{msg}");
     }
@@ -274,6 +276,121 @@ pub fn run(argv: &[&str]) -> CliResult<()> {
     Ok(())
 }
 
+fn split_invalid_records(
+    rconfig: &Config,
+    valid_flags: &[bool],
+    headers: &ByteRecord,
+    input_path: &str,
+    valid_suffix: &str,
+    invalid_suffix: &str,
+) -> CliResult<()> {
+    // track how many rows read for splitting into valid/invalid
+    // should not exceed row_number when aborted early due to fail-fast
+    let mut split_row_num: usize = 0;
+
+    // prepare output writers
+    let mut valid_wtr = Config::new(&Some(input_path.to_owned() + "." + valid_suffix)).writer()?;
+    valid_wtr.write_byte_record(headers)?;
+
+    let mut invalid_wtr =
+        Config::new(&Some(input_path.to_owned() + "." + invalid_suffix)).writer()?;
+    invalid_wtr.write_byte_record(headers)?;
+
+    let mut rdr = rconfig.reader()?;
+
+    let mut record = csv::ByteRecord::new();
+    while rdr.read_byte_record(&mut record)? {
+        split_row_num += 1;
+
+        // length of valid_flags is max number of rows we can split
+        if split_row_num > valid_flags.len() {
+            break;
+        }
+
+        // vector is 0-based, row_num is 1-based
+        let is_valid = valid_flags[split_row_num - 1];
+
+        if is_valid {
+            valid_wtr.write_byte_record(&record)?;
+        } else {
+            invalid_wtr.write_byte_record(&record)?;
+        }
+    }
+
+    valid_wtr.flush()?;
+    invalid_wtr.flush()?;
+
+    Ok(())
+}
+
+fn write_error_report(input_path: &str, validation_error_messages: Vec<String>) -> CliResult<()> {
+    let wtr_capacitys = env::var("QSV_WTR_BUFFER_CAPACITY")
+        .unwrap_or_else(|_| DEFAULT_WTR_BUFFER_CAPACITY.to_string());
+    let wtr_buffer_size: usize = wtr_capacitys.parse().unwrap_or(DEFAULT_WTR_BUFFER_CAPACITY);
+
+    let output_file = File::create(input_path.to_owned() + ".validation-errors.jsonl")?;
+
+    let mut output_writer = BufWriter::with_capacity(wtr_buffer_size, output_file);
+
+    // write out error report
+    for error_msg in validation_error_messages {
+        output_writer.write_all(error_msg.as_bytes())?;
+        // since writer is buffered, it's more efficient to do additional write than append Newline to message
+        output_writer.write_all(&[b'\n'])?;
+    }
+
+    // flush error report; file gets closed automagically when out-of-scope
+    output_writer.flush()?;
+
+    Ok(())
+}
+
+/// if given record is valid, CliResult would hold None, otherwise, Some(Value)
+fn do_json_validation(
+    headers: &ByteRecord,
+    record: &ByteRecord,
+    schema_json: &Value,
+    schema_compiled: &JSONSchema,
+) -> CliResult<Option<String>> {
+    // row number was added as last column
+    let row_number_string = str::from_utf8(record.get(headers.len()).unwrap()).unwrap();
+    let row_number: usize = row_number_string.parse::<usize>().unwrap();
+
+    let instance: Value = match to_json_instance(headers, record, schema_json) {
+        Ok(obj) => obj,
+        Err(e) => {
+            return fail!(format!(
+                "Unable to convert CSV to json. row: {row_number}, error: {e}"
+            ));
+        }
+    };
+
+    // debug!("instance[{row_number}]: {instance:?}");
+
+    match validate_json_instance(&instance, schema_compiled) {
+        Ok(mut validation_result) => {
+            let is_valid = validation_result.get("valid").unwrap().as_bool().unwrap();
+
+            if is_valid {
+                Ok(None)
+            } else {
+                // debug!("row[{row_number}] is invalid");
+
+                // insert row index and return validation error
+                validation_result.as_object_mut().unwrap().insert(
+                    "row_number".to_string(),
+                    Value::Number(Number::from(row_number)),
+                );
+
+                Ok(Some(validation_result.to_string()))
+            }
+        }
+        Err(e) => {
+            return fail!(format!("Unable to validate. row: {row_number}, error: {e}"));
+        }
+    }
+}
+
 /// convert CSV Record into JSON instance by referencing Type from Schema
 fn to_json_instance(headers: &ByteRecord, record: &ByteRecord, schema: &Value) -> Result<Value> {
     // make sure schema has expected structure
